@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2012, 2016, 2017 IBM Corporation and others.
+ * Copyright (c) 2012, 2017 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials are made 
  * available under the terms of the Eclipse Public License v1.0 
  * (http://www.eclipse.org/legal/epl-v10.html), and the Eclipse Distribution 
@@ -9,7 +9,7 @@
  *     IBM Corporation - initial API and implementation
  *******************************************************************************/
 /*eslint-env node */
-var api = require('../api'), writeError = api.writeError;
+var api = require('../api'), writeError = api.writeError, writeResponse = api.writeResponse;
 var path = require('path');
 var diff = require("./diff");
 var mTags = require("./tags");
@@ -29,18 +29,19 @@ module.exports = {};
 
 module.exports.router = function(options) {
 	var fileRoot = options.fileRoot;
-	if (!fileRoot) { throw new Error('options.root is required'); }
-	var contextPath = options.options.configParams["orion.context.path"] || "";
-
+	var gitRoot = options.gitRoot;
+	if (!fileRoot) { throw new Error('options.fileRoot is required'); }
+	if (!gitRoot) { throw new Error('options.gitRoot is required'); }
+	
 	module.exports.commitJSON = commitJSON;
 	module.exports.getDiff = getDiff;
 	module.exports.getCommitParents = getCommitParents;
 
 	return express.Router()
 	.use(bodyParser.json())
-	.get('/:scope'+ contextPath + '/file*', getCommit)
-	.put('/:commit'+ contextPath + '/file*', putCommit)
-	.post('/:commit'+ contextPath + '/file*', postCommit);
+	.get('/:scope'+ fileRoot + '*', getCommit)
+	.put('/:commit'+ fileRoot + '*', putCommit)
+	.post('/:commit'+ fileRoot + '*', postCommit);
 
 function commitJSON(commit, fileDir, diffs, parents) {
 	return {
@@ -50,10 +51,10 @@ function commitJSON(commit, fileDir, diffs, parents) {
 		"Children":[],
 		"CommitterEmail": commit.committer().email(),
 		"CommitterName": commit.committer().name(),
-		"ContentLocation": contextPath + "/gitapi/commit/" + commit.sha() + fileDir + "?parts=body",
-		"DiffLocation": contextPath + "/gitapi/diff/" + commit.sha() + fileDir,
-		"Location": contextPath + "/gitapi/commit/" + commit.sha() + fileDir,
-		"CloneLocation": contextPath + "/gitapi/clone" + fileDir,
+		"ContentLocation": {pathname: gitRoot + "/commit/" + commit.sha() + fileDir, query: {parts: "body"}},
+		"DiffLocation": gitRoot + "/diff/" + commit.sha() + fileDir,
+		"Location": gitRoot + "/commit/" + commit.sha() + fileDir,
+		"CloneLocation": gitRoot + "/clone" + fileDir,
 		"Diffs": diffs,
 		"Parents": parents,
 		"Message": commit.message(),
@@ -99,8 +100,136 @@ function getCommitLog(req, res) {
 		if (toDateFilter && commit.timeMs() > toDateFilter) return true;
 		return false;
 	}
+	function filterCommitPath(commit, fileDir, keep, ignore) {
+		// check if this commit is from a branch that should be ignored
+		for (var i = 0; i < ignore.length; i++) {
+			if (ignore[i].equal(commit.id())) {
+				for (var j = 0; j < commit.parentcount(); j++) {
+					// check if this parent should be skipped or iterated over
+					var shouldKeep = keep.some(function(oid) {
+						return oid.equal(commit.parentId(j));
+					});
+					if (!shouldKeep) {
+						ignore.push(commit.parentId(j));
+					}
+				}
+				return Promise.resolve(true);
+			}
+		}
+
+		var entryId;
+		return commit.getEntry(filterPath)
+		.then(function(entry) {
+			entryId = entry.id();
+			switch (commit.parentcount()) {
+				case 0:
+					// no parent then it was an ADD
+					return false;
+				case 1:
+					return commit.getParents(1)
+					.then(function(parents) {
+						return parents[0].getEntry(filterPath);
+					})
+					.then(function(parentEntry) {
+						// check if the file content changed
+						return entryId.equal(parentEntry.id()) !== 0;
+					})
+					.catch(function(err) {
+						if (err.toString().indexOf("does not exist in the given tree") !== -1) {
+							// not found in parent, was an ADD
+							return false;
+						}
+						throw err;
+					});
+				default:
+					var parentOids = [];
+					for (var i = 0; i < commit.parentcount(); i++) {
+						parentOids.push(commit.parentId(i));
+					}
+					var parents;
+					return commit.getParents(commit.parentcount())
+					.then(function(_parents) {
+						parents = _parents;
+						// for every parent, check if the entry exists
+						var promises = parents.map(function(parent) {
+							return parent.getEntry(filterPath).catch(function(err) {
+								if (err.toString().indexOf("does not exist in the given tree") !== -1) {
+									return null;
+								}
+								throw err;
+							});
+						});
+						return Promise.all(promises)
+					})
+					.then(function(entries) {
+						// find the common ancestor amongst the parents
+						return resolveAncestor(parentOids)
+						.then(function(base) {
+							for (var i = 0; i < entries.length; i++) {
+								if (entries[i] !== null && entries[i].id().equal(entryId)) {
+									// if the entry id is the same as a parent, that means
+									// the content from the other branches should be ignored
+									for (var j = 0; j < entries.length; j++) {
+										if (i !== j) {
+											ignore.push(parents[j].id());
+										}
+									}
+									// mark the common ancestor as not to be skipped over
+									if (base !== null) {
+										keep.push(base);
+									}
+									return true;
+								}
+							}
+							return  false;
+						});
+					});
+			}
+		})
+		.catch(function(err) {
+			if (err.toString().indexOf("does not exist in the given tree") !== -1) {
+				// doesn't exist, check if it was deleted
+				if (commit.parentcount() === 0) {
+					// no parent then definitely doesn't exist
+					return true;
+				}
+
+				return commit.getParents(1)
+				.then(function(parents) {
+					return parents[0].getEntry(filterPath);
+				})
+				.then(function(parentEntry) {
+					// if parent entry exists then the entry was deleted
+					return false;
+				})
+				.catch(function(err) {
+					return true;
+				});
+			}
+			throw err;
+		});
+	}
+	function resolveAncestor(parents) {
+		var tree = parents.pop();
+		var tree2 = parents.pop();
+		return git.Merge.base(repo, tree, tree2).then(function(base) {
+			if (parents.length === 0) {
+				return base;
+			} else {
+				parents.push(base);
+				return resolveAncestor(parents);
+			}
+		})
+		.catch(function(err) {
+			if (err.message === "No merge base found") {
+				// two commits in unrelated histories
+				return null;
+			}
+			throw err;
+		});
+	}
 	var commits = []	, repo;
-	function writeResponse(over) {
+	function sendResponse(over) {
 		var refs = scope.split("..");
 		var toRef, fromRef; 
 		if(refs.length === 1){
@@ -114,8 +243,8 @@ function getCommitLog(req, res) {
 			"Children": commits,
 			"RepositoryPath": filterPath,
 			"Type": "Commit",
-			"Location":contextPath + "/gitapi/commit/" + util.encodeURIComponent(scope) + contextPath + "/file" + req.params[0],
-			"CloneLocation": contextPath + "/gitapi/clone" + fileDir
+			"Location":gitRoot + "/commit/" + util.encodeURIComponent(scope) + fileRoot + req.params[0],
+			"CloneLocation": gitRoot + "/clone" + fileDir
 		};
 		if(mergeBase){
 			resp["AheadCount"] = aheadCount;
@@ -232,44 +361,87 @@ function getCommitLog(req, res) {
 		}
 
 		setupRevWalk(function() {
-			if (filterPath) {
-				fileHistoryWalk();
-			} else {
-				walk();
-			}
+			walk();
 		});
 
 		var count = 0;
+		var keep = [];
+		var ignore = [];
+		var commitJSONs = {};
 		filterPath = clone.getfileRelativePath(repo,req); 
 		function walk() {
 			return revWalk.next()
 			.then(function(oid) {
 				if (!oid) {
-					writeResponse(true);
+					sendResponse(true);
 					return;
 				}
 				return repo.getCommit(oid)
 				.then(function(commit) {
-					function applyFilter() {
-						if (filterCommit(commit) || page && count++ < skipCount) {//skip pages
+					function applyFilter(filter) {
+						if (filter || filterCommit(commit) || page && count++ < skipCount) {//skip pages
+							var keys = Object.keys(commitJSONs);
+							for (var i = 0 ; i < keys.length; i++) {
+								for (var j = 0; j < commitJSONs[keys[i]].Parents.length; j++) {
+									if (commitJSONs[keys[i]].Parents[j].Name === commit.sha() && commit.parentcount() !== 0) {
+										commitJSONs[keys[i]].Parents[j] = createParentJSON(commit.parentId(0).toString(), fileDir);
+										return walk();
+									}
+								}
+							}
 							return walk();
 						}
 						return Promise.all([getDiff(repo, commit, fileDir), getCommitParents(repo, commit, fileDir)])
 						.then(function(stuff) {
-							commits.push(commitJSON(commit, fileDir, stuff[0], stuff[1]));
+							var json = commitJSON(commit, fileDir, stuff[0], stuff[1]);
+							commitJSONs[commit.sha()] = json;
+							commits.push(json);
 							if (pageSize && commits.length === pageSize) {//page done
-								writeResponse();
+								if (filterPath) {
+									var missing = [];
+									// look for commits that are pointing at a parent that
+									// doesn't exist in the list of returned commits
+									var keys = Object.keys(commitJSONs);
+									for (var i = 0 ; i < keys.length; i++) {
+										for (var j = 0; j < commitJSONs[keys[i]].Parents.length; j++) {
+											var name = commitJSONs[keys[i]].Parents[j].Name;
+											if (commitJSONs[name] === undefined) {
+												// pointing at a commit that hasn't been resolved, mark it
+												missing.push(name);
+											}
+										}
+									}
+
+									if (missing.length === 0) {
+										sendResponse();
+									} else {
+										// resolve the parents so we actually have the JSON point
+										// at the true parent commit related to the file
+										var current = missing[0];
+										revWalk = repo.createRevWalk();
+										revWalk.sorting(git.Revwalk.SORT.TOPOLOGICAL);
+										revWalk.push(missing[0]);
+										resolveParents(missing, keep, ignore);
+									}
+								} else {
+									sendResponse();
+								}
 								return;
 							}
 							walk();
 						});
 					}
-					applyFilter();
+
+					if (filterPath) {
+						return filterCommitPath(commit, fileDir, keep, ignore).then(applyFilter);
+					} else {
+						return applyFilter(false);
+					}
 				});
 			})
 			.catch(function(error) {
 				if (error.errno === git.Error.CODE.ITEROVER) {
-					writeResponse(true);
+					sendResponse(true);
 				} else {
 					task.done({
 						HttpCode: 404,
@@ -283,71 +455,60 @@ function getCommitLog(req, res) {
 			});
 		}
 
-		function fileHistoryWalk() {
-			revWalk.getCommitsUntil(function(commit) {
-				// iterate through the entire history until we have no more commits
-				var parentId = commit.parentId(0);
-				return parentId !== null && parentId !== undefined;
-			}).then(function(array) {
-				setupRevWalk(function() {
-					// run a full history walk on the desired file
-					revWalk.fileHistoryWalk(filterPath, array.length).then(function(fileCommits) {
-						// client trying to skip more commits than that which affects the specified file,
-						// respond that we're done and there is nothing left
-						if (skipCount >= fileCommits.length) {
-							writeResponse(true);
-							return;
+		function resolveParents(missing, keep, ignore) {
+			var candidate;
+			return revWalk.next().then(function(oid) {
+				return repo.getCommit(oid);
+			})
+			.then(function(commit) {
+				candidate = commit;
+				return filterCommitPath(commit, fileDir, keep, ignore);
+			})
+			.then(function(filter) {
+				if (filter) {
+					// have to filter this commit, keep going
+					return resolveParents(missing, keep, ignore);
+				} else {
+					// this commit is not being filtered, update the parent information
+					var keys = Object.keys(commitJSONs);
+					for (var i = 0 ; i < keys.length; i++) {
+						for (var j = 0; j < commitJSONs[keys[i]].Parents.length; j++) {
+							if (commitJSONs[keys[i]].Parents[j].Name === missing[0]) {
+								commitJSONs[keys[i]].Parents[j] = createParentJSON(candidate.sha(), fileDir);
+							}
 						}
+					}
 
-						// remove the elements that need to be skipped over
-						fileCommits.splice(0, skipCount);
-						var done = true;
-						if (fileCommits.length > pageSize) {
-							// we have more commits than the page size, we won't be done
-							done = false;
-							// remove the extra commits as they don't need to be processed
-							fileCommits.splice(pageSize, fileCommits.length - pageSize);
-						}
-						var fullCommits = [];
-						function getFullCommits(fileCommits, idx) {
-							repo.getCommit(fileCommits[idx].commit.id()).then(function(c) {
-								fullCommits.push(c);
-
-								// when we've retrieved everything, iterate over them
-								if (fullCommits.length === fileCommits.length) {
-									processCommits(fullCommits, 0);	
-								} else {
-									// get the full commit of the next commit
-									getFullCommits(fileCommits, idx + 1);
-								}
-							});
-						}
-						// the commit objects from fileHistoryWalk(*) fail on getTree()
-						// which prevents us from grabbing a diff,
-						// grab a full commit from the repo using the oid
-						getFullCommits(fileCommits, 0);
-
-						function processCommits(fullCommits, idx) {
-							var commit = fullCommits[idx];
-							return Promise.all(
-								[ getDiff(repo, commit, fileDir), getCommitParents(repo, commit, fileDir) ]
-							).then(function(stuff) {
-								commits.push(commitJSON(commit, fileDir, stuff[0], stuff[1]));
-								if (idx === fullCommits.length - 1) {
-									// we're at the end, stop iterating
-									// and write the response
-									writeResponse(done);
-									return;
-								}
-								processCommits(fullCommits, idx + 1);
-							});
-						}
+					// remove the resolved commit
+					missing.shift();
+					if (missing.length === 0) {
+						sendResponse();
+					} else {
+						// still more to go, resolve the next one
+						revWalk = repo.createRevWalk();
+						revWalk.sorting(git.Revwalk.SORT.TOPOLOGICAL);
+						revWalk.push(missing[0]);
+						return resolveParents(missing, keep, ignore);
+					}
+				}
+			})
+			.catch(function(error) {
+				if (error.errno === git.Error.CODE.ITEROVER) {
+					sendResponse(true);
+				} else {
+					task.done({
+						HttpCode: 404,
+						Code: 0,
+						DetailedMessage: error.message,
+						JsonData: {},
+						Message: error.message,
+						Severity: "Error"
 					});
-				});
+				}
 			});
 		}
 	}
-	
+
 	clone.getRepo(req)
 	.then(function(_repo) {
 		repo = _repo;
@@ -368,19 +529,29 @@ function getCommitLog(req, res) {
 			})
 			.then(function(commit) {
 				commit1 = commit;
-				return git.Merge.base(repo, commit0, commit1).then(function(oid){
-					return oid;
+				// find the common ancestor for calculating incoming/outgoing changes
+				return git.Merge.base(repo, commit0, commit1)
+				.catch(function(err) {
+					if (err.message === "No merge base found") {
+						return null;
+					}
+					throw err;
 				});
 			})
 			.then(function(oid) {
-				mergeBaseCommitOid = oid;
 				var revWalkForCounting1 = repo.createRevWalk();
-				revWalkForCounting1.sorting(git.Revwalk.SORT.TOPOLOGICAL);
-				revWalkForCounting1.pushRange(mergeBaseCommitOid.tostrS() + ".." + commit0.id().tostrS());	//count for incomings
-				
 				var revWalkForCounting2 = repo.createRevWalk();
+				revWalkForCounting1.sorting(git.Revwalk.SORT.TOPOLOGICAL);
 				revWalkForCounting2.sorting(git.Revwalk.SORT.TOPOLOGICAL);
-				revWalkForCounting2.pushRange(mergeBaseCommitOid.tostrS() + ".." + commit1.id().tostrS()); // count for outgoing
+				if (oid === null) {
+					// no common ancestor, just count everything
+					revWalkForCounting1.push(commit0.id().tostrS()); // incoming changes
+					revWalkForCounting2.push(commit1.id().tostrS()); // outgoing changes
+				} else {
+					mergeBaseCommitOid = oid;
+					revWalkForCounting1.pushRange(mergeBaseCommitOid.tostrS() + ".." + commit0.id().tostrS()); // incoming changes
+					revWalkForCounting2.pushRange(mergeBaseCommitOid.tostrS() + ".." + commit1.id().tostrS()); // outgoing changes
+				}
 				
 				return Promise.all([countCommit(revWalkForCounting1),countCommit(revWalkForCounting2)]);	
 			}).then(function(results){
@@ -389,7 +560,7 @@ function getCommitLog(req, res) {
 				if(mergeBaseCommitOid){
 					log(repo, mergeBaseCommitOid.tostrS());
 				} else {
-					writeResponse(true);
+					sendResponse(true);
 				}
 			});
 		} else {
@@ -419,14 +590,18 @@ function getCommitParents(repo, commit, fileDir) {
 	return commit.getParents()
 	.then(function(parents) {
 		return parents.map(function(parent) {
-			return {
-				"Location": contextPath + "/gitapi/commit/" + parent.sha() + fileDir,
-				"Name": parent.sha()
-			};
+			return createParentJSON(parent.sha(), fileDir);
 		});
 	});
 }
 
+function createParentJSON(sha, fileDir) {
+	return {
+		"Location": gitRoot + "/commit/" + sha + fileDir,
+		"Name": sha
+	};
+}
+
 function getCommitRefs(repo, fileDir, commits) {
 	return new Promise(function (fulfill){
 		if (!commits.length) return fulfill();
@@ -443,8 +618,11 @@ function getCommitRefs(repo, fileDir, commits) {
 					var id,commit;
 					var shortName = ref.replace("refs/tags/", "").replace("refs/remotes/", "").replace("refs/heads/", "");
 					if (ref.indexOf("refs/tags/") === 0) {
+						var annotated = false;
 						return git.Tag.lookup(repo,oid).then(function(tag){
 							id = tag.targetId();
+							// tag lookup succeeded, it's an annotated tag then
+							annotated = true;
 						}).catch(function(){
 							id = oid.toString();
 						})
@@ -452,7 +630,7 @@ function getCommitRefs(repo, fileDir, commits) {
 							commit = map[id];
 							if (commit) {
 								var tags = commit.Tags || (commit.Tags = []);
-								tags.push(mTags.tagJSON(fullName, shortName, id, undefined, fileDir));
+								tags.push(mTags.tagJSON(fullName, shortName, id, undefined, fileDir, annotated));
 							}
 							cb();
 						});
@@ -515,7 +693,7 @@ function getDiff(repo, commit, fileDir) {
 			return {
 				"ChangeType": type,
 				"ContentLocation": fileDir + "/" + path1,
-				"DiffLocation": contextPath + "/gitapi/diff/" + range + fileDir + "/" + path1,
+				"DiffLocation": gitRoot + "/diff/" + range + fileDir + "/" + path1,
 				"NewPath": newFilePath,
 				"OldPath": oldFilePath,
 				"Type": "Diff"
@@ -527,7 +705,7 @@ function getDiff(repo, commit, fileDir) {
 			"Children": diffs
 		};
 		if (patches.length > 100) {
-			result.NextLocation = contextPath + "/gitapi/diff/" + range + fileDir + "?pageSize=" + pageSize + "&page=" + (page + 1);
+			result.NextLocation = {pathname: gitRoot + "/diff/" + range + fileDir, query: {pageSize: pageSize, page: page + 1}};
 		}
 		return result;
 	});
@@ -554,9 +732,7 @@ function getCommitBody(req, res) {
 	})
 	.then(function(blob) {
 		var resp = blob.toString();
-		res.setHeader('Content-Type', 'application/octect-stream');
-		res.setHeader('Content-Length', resp.length);
-		res.status(200).end(resp);
+		writeResponse(200, res, {'Content-Type':'application/octect-stream','Content-Length': resp.length}, resp, false, true);
 	}).catch(function(err) {
 		writeError(404, res, err.message);
 	});
@@ -565,12 +741,12 @@ function getCommitBody(req, res) {
 function identifyNewCommitResource(req, res, newCommit) {
 	var originalUrl = url.parse(req.originalUrl, true);
 	var segments = originalUrl.pathname.split("/");
-	var contextPathSegCount = contextPath.split("/").length - 1;
+	var contextPathSegCount = req.contextPath.split("/").length - 1;
 	segments[3 + contextPathSegCount] = segments[3 + contextPathSegCount] + ".." + util.encodeURIComponent(newCommit);
 	var location = url.format({pathname: segments.join("/"), query: originalUrl.query});
-	res.status(200).json({
+	writeResponse(200, res, null, {
 		"Location": location
-	});
+	}, false); // Avoid double encoding
 }
 
 function revert(req, res, commitToRevert) {
@@ -594,7 +770,7 @@ function revert(req, res, commitToRevert) {
 		return theRepo.stateCleanup();
 	})
 	.then(function() {
-		res.status(200).json({
+		writeResponse(200, res, null, {
 			"Result": theRC ? "FAILURE" : "OK"
 		});
 	})
@@ -635,7 +811,7 @@ function cherryPick(req, res, commitToCherrypick) {
 		return git.Reference.nameToId(theRepo, "HEAD");
 	})
 	.then(function(newHead) {
-		res.status(200).json({
+		writeResponse(200, res, null, {
 			"Result": theRC ? "FAILED" : "OK",
 			"HeadUpdated": !theRC && theHead !== newHead
 		});
@@ -647,7 +823,7 @@ function cherryPick(req, res, commitToCherrypick) {
 	})
 	.catch(function(err) {
 		if(err.message.indexOf("Cannot create a tree") !== -1){
-			res.status(200).json({
+			writeResponse(200, res, null, {
 				"HeadUpdated": true,
 				"Result": "CONFLICTING"
 			});
@@ -697,7 +873,26 @@ function rebase(req, res, commitToRebase, rebaseOperation) {
 				break;
 				
 			case "SKIP":
-				throw new Error("Not implemented yet");
+				return git.Checkout.head(repo, {
+					checkoutStrategy:
+						git.Checkout.STRATEGY.FORCE |
+						git.Checkout.STRATEGY.RECREATE_MISSING |
+						git.Checkout.STRATEGY.REMOVE_UNTRACKED
+				})
+				.then(function() {
+					return git.Rebase.open(repo, {});
+				}).then(function(rebase) {
+					if (rebase.operationCurrent() === rebase.operationEntrycount() - 1) {
+						// if skipping the last operation, then we're done here
+						rebase.finish(repo.defaultSignature());
+						// return the commit that we're currently on
+						return head;
+					}
+					// move to the next one and continue
+					return rebase.next().then(function(rebaseOperation) {
+						return repo.continueRebase();
+					});
+				});
 				
 			default:
 				work = repo.rebaseBranches("HEAD", commitToRebase, null, null, null);
@@ -735,7 +930,7 @@ function rebase(req, res, commitToRebase, rebaseOperation) {
 		} else if (rebaseOperation === "ABORT") {
 			rebaseResult = "ABORTED";
 		}
-		res.status(200).json({
+		writeResponse(200, res, null, {
 			"Result": rebaseResult
 		});
 	})
@@ -745,8 +940,9 @@ function rebase(req, res, commitToRebase, rebaseOperation) {
 
 }
 
-function merge(req, res, commitToMerge, squash) {
+function merge(req, res, branchToMerge, squash) {
 	var repo, head, commit, oid, paths;
+	var conflicting = false;
 	clone.getRepo(req)
 	.then(function(_repo) {
 		repo = _repo;
@@ -754,39 +950,96 @@ function merge(req, res, commitToMerge, squash) {
 	})
 	.then(function(_head) {
 		head = _head;
-		return repo.getReferenceCommit(commitToMerge);
+		return repo.getReferenceCommit(branchToMerge);
 	})
 	.then(function(_commit) {
 		commit = _commit;
 		if (squash) {
-			throw new Error("Not implemented yet");
-		}
-		return repo.mergeBranches("HEAD", commitToMerge)
-		.then(function(_oid) {
-			oid = _oid;
-		})
-		.catch(function(index) {
-			paths = {};
-			index.entries().forEach(function(entry) {
-				if (git.Index.entryIsConflict(entry)) {
-					paths[entry.path] = "";
+			// find a common ancestor
+			return git.Merge.base(repo, head, commit.id()).then(function(base) {
+				// the common ancestor is the commit we're trying to merge
+				if (base.equal(commit.id())) {
+					// already up-to-date, nothing to do
+					oid = head;
+					return;
 				}
+
+				var mergeIndex;
+				// try to merge
+				return git.Merge.commits(repo, head, commit)
+				.then(function(index) {
+					mergeIndex = index;
+					return repo.getStatus();
+				})
+				.then(function(statuses) {
+					paths = {};
+					// see what's in the index/wd
+					statuses = statuses.map(function(status) {
+						return status.path();
+					});
+					// compare it with what's in the index
+					var entries = mergeIndex.entries();
+					entries.forEach(function(entry) {
+						if (statuses.indexOf(entry.path) !== -1) {
+							paths[entry.path] = "";
+						}
+					});
+					// no conflicts, then we're fine
+					if (Object.keys(paths).length === 0) {
+						// common ancestor is HEAD, fast-forward
+						if (base.equal(head.id())) {
+							oid = commit.id();
+						}
+
+						conflicting = mergeIndex.hasConflicts();
+						paths = undefined;
+						// checkout the merged index
+						return git.Checkout.index(repo, mergeIndex, {
+							checkoutStrategy: git.Checkout.STRATEGY.FORCE
+						});
+					}
+				});
 			});
-			return git.Checkout.index(repo, index, {
-				checkoutStrategy: git.Checkout.STRATEGY.ALLOW_CONFLICTS,
-				ourLabel: "HEAD",
-				theirLabel: commit.sha()
+		}
+
+		// try to find a common ancestor before merging
+		return git.Merge.base(repo, head, commit.id())
+		.then(function() {
+			return repo.mergeBranches("HEAD", branchToMerge)
+			.then(function(_oid) {
+				oid = _oid;
+			})
+			.catch(function(index) {
+				// the merge failed due to conflicts
+				conflicting = true;
+				return forceMerge(repo, head, commit, branchToMerge, false, function(conflictingPaths) {
+					paths = conflictingPaths;
+				});
 			});
+		})
+		.catch(function(err) {
+			if (err.message === "No merge base found") {
+				// no common ancestor between the two branches, force the merge
+				return forceMerge(repo, head, commit, branchToMerge, true, function(conflictingPaths) {
+					paths = conflictingPaths;
+				});
+			}
+			throw err;
 		});
 	})
 	.then(function() {
-		var mergeResult = "MERGED";
-		if (!paths) {
-			if (oid && oid.toString() === head.id().toString()) mergeResult = "ALREADY_UP_TO_DATE";
-			else if (oid && oid.toString() === commit.id().toString()) mergeResult = "FAST_FORWARD";
+		var mergeResult = squash ? "MERGED_SQUASHED" : "MERGED";
+		if (oid && oid.toString() === head.id().toString()) {
+			mergeResult = "ALREADY_UP_TO_DATE";
+		} else if (oid && oid.toString() === commit.id().toString()) {
+			mergeResult = squash ? "FAST_FORWARD_SQUASHED" : "FAST_FORWARD";
+		} else if (paths) {
+			mergeResult = "FAILED";
+		} else if (conflicting) {
+			mergeResult = "CONFLICTING";
 		}
-		res.status(200).json({
-			"Result": paths ? "CONFLICTING" : mergeResult,
+		writeResponse(200, res, null, {
+			"Result": mergeResult,
 			"FailingPaths": paths
 		});
 	})
@@ -795,8 +1048,97 @@ function merge(req, res, commitToMerge, squash) {
 	});
 }
 
+/**
+ * Force a merge from the specified comment onto HEAD. This may leave
+ * the repository in a conflicted state. If there are no conflits, a
+ * merge commit will be created to complete the merge.
+ * 
+ * @param {Repository} repo the repository to perform the merge in
+ * @param {Commit} head the commit that HEAD is pointing at
+ * @param {Commit} commit the commit to merge in
+ * @param {String} branchToMerge the name of the other branch
+ * @param {boolean} createMergeCommit <tt>true</tt> if a merge commit should be created automatically,
+ *                                    <tt>false</tt> otherwise
+ * @param {Function} conflictingPathsCallback the callback to notify if the merge fails due to
+ *                                            conflicting paths in the working directory or the index
+ */
+function forceMerge(repo, head, commit, branchToMerge, createMergeCommit, conflictingPathsCallback) {
+	return git.AnnotatedCommit.lookup(repo, commit.id())
+	.then(function(annotated) {
+		var retCode = git.Merge.merge(repo, annotated, null, null);
+		if (retCode === git.Error.CODE.ECONFLICT) {
+			// checkout failed due to a conflict
+			return getConflictingPaths(repo, head, commit).then(conflictingPathsCallback);
+		} else if (retCode !== git.Error.CODE.OK) {
+			throw new Error("Internal merge failure (error code " + retCode + ")");
+		}
+
+		if (createMergeCommit) {
+			var signature = repo.defaultSignature();
+			var message = "Merged branch '" + branchToMerge + "'"; 
+			return createCommit(repo,
+				signature.name(), signature.email(),
+				signature.name(), signature.email(),
+				message, false, false);
+		}
+	});
+}
+
+function getConflictingPaths(repo, head, commit) {
+	var tree2;
+	var diffPaths = [];
+	return commit.getTree()
+	.then(function(tree) {
+		tree2 = tree;
+		// get the common ancestor between HEAD and the other branch
+		return git.Merge.base(repo, head, commit)
+		.then(function(ancestor) {
+			return ancestor.getTree();
+		})
+		.catch(function() {
+			// failed to find common ancestor, use HEAD tree instead
+			return head.getTree();
+		});
+	})
+	.then(function(tree) {
+		// diff the two trees
+		return git.Diff.treeToTree(repo, tree, tree2, null)
+	})
+	.then(function(diff) {
+		return diff.patches();
+	})
+	.then(function(patches) {
+		// get the path of every modified file
+		patches.forEach(function(patch) {
+			diffPaths.push(patch.newFile().path());
+		});
+		return repo.getStatusExt({
+			flags: git.Status.OPT.INCLUDE_UNTRACKED |
+				git.Status.OPT.RECURSE_UNTRACKED_DIRS
+		});
+	})
+	.then(function(statuses) {
+		var paths = {};
+		// check what's in the index and wd
+		statuses.forEach(function(file) {
+			var path = file.path();
+			if (file.statusBit() !== 0 && diffPaths.indexOf(path) !== -1) {
+				// modified file in the index/wd conflicts with the merge
+				paths[path] = "";
+			}
+		});
+		return paths;
+	});
+}
+
 function createCommit(repo, committerName, committerEmail, authorName, authorEmail, message, amend, insertChangeid){
 	var index, oid, author, committer;
+	var state = repo.state();
+	var mergeRequired = state === git.Repository.STATE.MERGE;
+	if (amend && mergeRequired) {
+		// libgit2's API doesn't allow this
+		throw new Error("Cannot amend a commit to have more merges");
+	}
 	return repo.index()
 	.then(function(indexResult) {
 		index = indexResult;
@@ -833,11 +1175,32 @@ function createCommit(repo, committerName, committerEmail, authorName, authorEma
 			}
 			if (amend) {
 				return parent.amend("HEAD",  author, committer, null, message, oid);
+			} else if (mergeRequired) {
+				var promises = [ Promise.resolve(parent) ];
+				// get merge heads
+				return repo.mergeheadForeach(function(oid) {
+					promises.push(repo.getCommit(oid));
+				})
+				.then(function() {
+					// wait for all parents to be resolved
+					return Promise.all(promises);
+				})
+				.then(function(parentCommits) {
+					// create the merge commit on top of the MERGE_HEAD parents
+					return repo.createCommit("HEAD", author, committer, message, oid, parentCommits);
+				});
 			}
 			return repo.createCommit("HEAD", author, committer, message, oid, [parent]);
 		});
 	})
 	.then(function(id) {
+		if (mergeRequired) {
+			// a merge commit was created, cleanup MERGE_* files in .git/
+			var retCode = repo.stateCleanup();
+			if (retCode !== git.Error.CODE.OK) {
+				throw new Error("Internal merge cleanup failed (error code " + retCode + ")");
+			}
+		}
 		return git.Commit.lookup(repo, id);
 	});
 }
@@ -867,13 +1230,10 @@ function tag(req, res, commitId, name, isAnnotated, message) {
 		return getCommitParents(theRepo, thisCommit, fileDir);
 	})
 	.then(function(parents){
-		theParents = parents;
+		writeResponse(200, res, null, commitJSON(thisCommit, fileDir, theDiffs, parents), true);
 	})
 	.catch(function(err) {
 		writeError(403, res, err.message);
-	})
-	.done(function() {
-		res.status(200).json(commitJSON(thisCommit, fileDir, theDiffs, theParents));
 	});
 }
 
@@ -925,7 +1285,7 @@ function postCommit(req, res) {
 	var isInsertChangeId = req.body.ChangeId;	
 
 	var theRepo, thisCommit;
-	var theDiffs = [], theParents;
+	var theDiffs = [];
 	
 	clone.getRepo(req)
 	.then(function(repo) {
@@ -944,13 +1304,10 @@ function postCommit(req, res) {
 		return getCommitParents(theRepo, thisCommit, fileDir);
 	})
 	.then(function(parents){
-		theParents = parents;
+		writeResponse(200, res, null, commitJSON(thisCommit, fileDir, theDiffs, parents), true);
 	})
 	.catch(function(err) {
 		writeError(403, res, err.message);
-	})
-	.done(function() {
-		res.status(200).json(commitJSON(thisCommit, fileDir, theDiffs, theParents));
 	});
 }
 
